import os
import sys
import numpy
from PIL import Image
import pickle

#check if all arguments are given
if(len(sys.argv)<6):
    print("COMMAND LINE ARGUMENTS ARE MISSING")
    exit()

i1 = 0
j1 = 0
#kernel function
vector_size = 625 #25 X 25 image will get converted to a 625 size vector.

# Function to calculate the centroid of negative and positive examples
def calculate_centroid(vectors):
            global vector_size
            centroid = numpy.zeros(vector_size)
            
            
            for vector in vectors:
                
                
                centroid = numpy.add(centroid, vector)
                
            centroid = numpy.divide(centroid, len(vectors))
            
            return centroid
            
#convert image to vector
    
def image2vector(img):
    imageArray = numpy.array(img.convert("L"))
    
    vector = imageArray.flatten().astype("float64")
    vector = vector/255
    return vector
    
#The kernel function of degree 4
def polynomial_kernel(vector1, vector2):
    vector1_transpose = numpy.transpose(vector1)
    product = numpy.dot(vector1_transpose, vector2)
    
    product = product + 1
    product = (product) **4
    return product      

def calculate_r_plusminus(i_plus, m_plus, is_plus_minus):
    max_norm_val = float('-inf')
    if(is_plus_minus == True):
        for vector in i_plus:
            
            #do subtraction
            
            difference = numpy.subtract(vector, m_plus)
            
            norm_val = numpy.linalg.norm(difference)
            
            if(norm_val > max_norm_val):
                max_norm_val = norm_val
        return max_norm_val
    else:
        vector = i_plus
        difference = numpy.subtract(vector, m_plus)
        norm_val = numpy.linalg.norm(difference)
        return norm_val
    
    
def compute_x_dash(x, lambda_value, centroid):
    
    x_temp = lambda_value * numpy.array(x)
    
    
    constant = (1 - lambda_value) * numpy.array(centroid)
    
    x_dash = numpy.add(x_temp, constant)
    return x_dash
    
def sk_alg_values(x_dash, alpha):
    
    global i1
    global j1
    
    #set the values for alpha i1 and alpha j1
    
    alpha[i1] = 1
    alpha[j1] = 1
        
    a = polynomial_kernel(x_dash[i1], x_dash[i1])
    
    b = polynomial_kernel(x_dash[j1], x_dash[j1])
    
    c = polynomial_kernel(x_dash[i1], x_dash[j1])
    
    
    d = []
    e = []
    
    for xi_dash in x_dash:
        
        d.append(polynomial_kernel(xi_dash,x_dash[i1]))
        e.append(polynomial_kernel(xi_dash, x_dash[j1]))
    
    
    
    return a, b, c, d, e, alpha
    
#calculating the argmin value to update t
def calculate_argmin(d_values, e_values, a, b, c):
        global i1
        global j1
        
        n = len(d_values)
        minimum_mi = float('inf')
        mi = 0
        t = 0
        common_base = a + b - 2 * c
        
        for i in range(0, n):
        
            if(i >= i1 and i < j1): #means the example is a positive example
                numerator = d_values[i] - e_values[i] + b - c 
                
                denominator = pow(common_base, 0.5)
                
                mi = numerator/denominator
                
            elif(i >= j1 and i < n):
                numerator = e_values[i] - d_values[i] + a - c
                denominator = pow(common_base, 0.5)
                mi = numerator/denominator
                if(mi < minimum_mi):
                    minimum_mi = mi
                    t = i
        
            
        return t, minimum_mi
        
        

def sk_wrapper(x_dash):
    
    #this is like a wrapper function that calls other sub modules like initialization, updation and termination.
    data_set_len = len(x_dash)
    alpha = [0] * data_set_len
    
    a, b, c, d, e, alpha = sk_alg_values(x_dash, alpha)
    steps = 0   
    t, argmin = calculate_argmin(d, e, a, b, c)
    
    #we will run a while loop here which checks the exit condition.
    while(check_exit_threshold(a, b, c, argmin) >= float(sys.argv[1]) and steps <= int(sys.argv[2])):
        #check if t lies in the range of positive examples or negative examples    
        #if the t lies in the range of positive examples then we call the positive update function.
        if(t >= i1 and t < j1):
            
            alpha, a, c, d = positive_update(x_dash, alpha, a, b, c, d, e, t)
            
            
        elif(t >= j1 and t < data_set_len):
            
            alpha, b, c, e = negative_update(x_dash, alpha, a, b, c, d, e, t)
            
        
        t, argmin = calculate_argmin(d,e,a,b,c)
        steps = steps + 1 
        
            
            
    return alpha, a, b        

#function to check exit conditions 
def check_exit_threshold(a, b, c, argmin):
    base = a + b - 2 * c
    val = pow(base, 0.5)
    
    return val - argmin
    
def positive_update(x_dash, alpha, a, b, c, d, e, t):
    dt = d[t]
    et = e[t]
    x_dash_t = x_dash[t]
    numerator = a - dt + et - c
    denominator = a + polynomial_kernel(x_dash[t], x_dash[t]) - 2*(dt - et)
    val = numerator/denominator
    q = min(1, val)
    
    one_minus_q = 1 - q
    q_square = q ** 2
    #update for alpha involves a loop
    
    for alpha_i in range(0, len(alpha)):
        alpha_one = one_minus_q * alpha[alpha_i]
        if(alpha_i == t):
            delta = 1
            
        else:
            delta = 0
        alpha_two = q * delta
        alpha[alpha_i] = alpha_one + alpha_two
        
    #update for A
    a1 = a * pow(one_minus_q, 2)
    a2 = 2 * one_minus_q*q * dt
    a3 = q_square * polynomial_kernel(x_dash_t, x_dash_t)
    a = a1 + a2 + a3
    
    #Update for C
    
    c1 = one_minus_q * c
    c2 = q * et
    c = c1 + c2
    
    #Update for D involves a loop
    for i in range(0, len(d)):
        d1 = one_minus_q * d[i]
        d2 = q * polynomial_kernel(x_dash[i], x_dash_t)
        d[i] = d1 + d2
    return alpha, a, c, d
    
    
#negative update function
def negative_update(x_dash, alpha, a, b, c, d, e, t):
    dt = d[t]
    et = e[t]
    et_dt = et-dt
    numerator = b - et + dt - c
    denominator = b + polynomial_kernel(x_dash[t], x_dash[t]) - 2* (et - dt)
    val = numerator/denominator
    q = min(1, val)
    
    one_minus_q = 1 - q
    q_square = q ** 2
    
    b1 = b * pow(one_minus_q, 2)
    b2 = 2 * (one_minus_q)*q * et
    b3 = q_square * polynomial_kernel(x_dash[t], x_dash[t])
    b = b1 + b2 + b3
    
    c1 = one_minus_q * c
    c2 = q * dt
    c = c1 + c2
    
    for alpha_i in range(0, len(alpha)):
        alpha_one = one_minus_q * alpha[alpha_i]
        if(alpha_i == t):
            delta = 1
            
        else:
            delta = 0
        alpha_two = q * delta
        alpha[alpha_i] = alpha_one + alpha_two
    
    for i in range(0, len(e)):
        e1 = one_minus_q * e[i]
        e2 = q * polynomial_kernel(x_dash[i], x_dash[t])
        e[i] = e1 + e2
    return alpha, b, c, e
    
    
#if data folder does not exist or is empty
if (os.path.isdir(sys.argv[5])==False) or(len(os.listdir(sys.argv[5])) == 0):
    print("NO DATA")

else:
    i_plus = []
    i_minus = []
  
    for filename in sorted(os.listdir(sys.argv[5])):
        img = Image.open(os.path.join(sys.argv[5], filename))
        
        vector = image2vector(img)
        
        if(sys.argv[3] in filename):
         i_plus.append(vector)
        else:
         i_minus.append(vector)

    
    #calculate m+ and m-
    m_plus = calculate_centroid(i_plus)
    
    m_minus = calculate_centroid(i_minus)
    
    
    #calculate r, r+ and r-
    
    r = calculate_r_plusminus(m_plus, m_minus, False)
    
    r_plus = calculate_r_plusminus(i_plus, m_plus, True)
    
    r_minus = calculate_r_plusminus(i_minus, m_minus, True)
    
    denom = r_plus + r_minus
    lambda_val = r/ denom
    
    
    #create the xplusdash set and xminusdash set using 2 for loops or you could write a function but I find this easier.
    
    x_plus_dash = []
    x_minus_dash = []
    for i in i_plus:
        x_plus_dash.append(compute_x_dash(i, lambda_val, m_plus))    
        
    for i in i_minus:
        x_minus_dash.append(compute_x_dash(i, lambda_val, m_minus))
    
    x_minus_len = len(x_minus_dash)
    x_plus_len = len(x_plus_dash)
    x_dash = []
    x_plus_dash.extend(x_minus_dash)
    x_dash.extend(x_plus_dash)
    
    x_dash_len = len(x_dash)
    
    j1 = x_dash_len - x_minus_len
    
    
    
    alpha, a, b = sk_wrapper(x_dash)
    
    #only extract the non zero alphas from the array
    
    support_vectors = []
    for i in range(0, len(alpha)):
        
        if(alpha[i] != 0):
            pair = []
            pair.append(i)
            pair.append(alpha[i])
            support_vectors.append(pair)
        
        
    
    
    Dict = {"Class": sys.argv[3],
           "centroid for positive data": m_plus,
           "centroid for negative data": m_minus,
           "lambda": lambda_val,
           "A": a,
           "B": b,
           "weights": support_vectors}
    
#dump the values in text file  
    with open(sys.argv[4], 'wb') as fp:
        pickle.dump(Dict, fp)
    
    
    
    
    


