import os, sys
from typing import final
import numpy 
import re
from PIL import Image, ImageDraw, ImageFont
import pickle

def polynomial_kernel(vector1, vector2):
    '''
    To calculate value of kernel with polynomial 4
    '''
    vector1_transpose = numpy.transpose(vector1)
    product = numpy.dot(vector1_transpose, vector2)
    product = numpy.add(product, 1)
    product = (product) **4
    return product 

def image2vector(img):
    '''
    To convert image to vector
    '''
    imageArray = numpy.array(img.convert("L"))
    vector = imageArray.flatten().astype("float64")
    vector = vector/255
    return vector
    
def start_test():
    '''
    Checking exit conditions , reading the text file generated from training and calculating
     the false positives , false negatives and correctly predicted values for our test data
    '''
    
    if(len(sys.argv)<4):
        print("COMMAND LINE ARGUMENTS ARE MISSING")
        exit()
    if(os.path.isdir(sys.argv[3])==False) or(len(os.listdir(sys.argv[3])) == 0):
        print("NO TEST DATA")
        exit()
    if (os.path.isdir(sys.argv[2])==False) or(len(os.listdir(sys.argv[2])) == 0):
        print("NO TRAINING DATA")
        exit()
    if (sys.argv[1]).endswith('.txt'):
        pass
    else:
        print("MODEL FILE IS NOT OF THE CORRECT FORMAT")
        exit()
    if (os.path.exists(sys.argv[1])):
        pass
    else:
        print("CAN'T FIND MODEL FILE")
        exit()

    false_positives,count,number_correct,false_negatives = 0,0,0,0
           
    with open(sys.argv[1],'rb') as fp:
        model = pickle.load(fp)
    res_list ={}
    for f in os.listdir(sys.argv[3]):
        class_value = re.match(r'(.*)_(.*).png',f)
        if str(class_value.group(2)) == model['Class']:            
            final_class = 1
        else:
            final_class = 0
        f_new = Image.open(os.path.join(sys.argv[3], f))        
        x_test = image2vector(f_new)
        g_val,class_predicted = implement_svm(x_test,model)
        img_no_initial=re.findall(r'[0-9]*',f)[0]
        img_no = int(img_no_initial)
        count = count + 1
        if int(final_class)-int(class_predicted) == 0:
            number_correct = number_correct + 1
            res_list[img_no] = 'Correct'
        elif int(final_class)-int(class_predicted) <= 0:
            false_positives = false_positives + 1
            res_list[img_no] = 'False Positive'
        else:
            false_negatives = false_negatives + 1  
            res_list[img_no] = 'False Negative'
    
    
    dictionary_items = res_list.items()
    sorted_items = sorted(dictionary_items)
    for k, v in sorted_items:
        print(k, v)
    
    print("Fraction correct: "+str((float(number_correct)/float(count))))
    print("Fraction False Positive: "+str(float(false_positives)/float(count)))
    print("Fraction False Negative: "+str(float(false_negatives)/float(count)))

def xdash_calculate(x1,m,lam):
    '''
    calculate new centroid
    '''
    new_centroid = (1.0-lam)*numpy.array(m)
    new_value = lam*numpy.array(x1)
    x1dash = numpy.add(new_value,new_centroid)
    return x1dash

def get_train_data(index,class_train,i_plus):
    '''
    Convert the required train data into vectors 
    '''
    y_res = -1
    class_value = re.match(r'(.*)_(.*).png',i_plus[index])
    if str(class_value.group(2)) == class_train:
            y_res = 1
    im = Image.open(os.path.join(sys.argv[2], i_plus[index]))
    new_arr = numpy.array(list(im.getdata()))
    return new_arr/ 255 , y_res        
    
def implement_svm(x_test,model):
    '''
    To Predict the result of the test image using svm kernel
    '''
    final_sum = 0.0
    i_plus = []
    i_minus = []
    for filename in os.listdir(sys.argv[2]):
        if(model["Class"] in filename):
            i_plus.append(filename)
        else:
            i_minus.append(filename)
    i_plus.extend(i_minus)
    for value in model["weights"]:
        xi,y_res = get_train_data(value[0],model["Class"],i_plus)
        if y_res < 0:
            xi_dash = xdash_calculate(xi,model["centroid for negative data"],float(model["lambda"]))
        else:
            xi_dash = xdash_calculate(xi,model["centroid for positive data"],float(model["lambda"]))
        
        final_sum = value[1]* y_res * polynomial_kernel(x_test,xi_dash)
    g = final_sum + (float(model["B"]) - float(model["A"]))/2
    # If Predicted value belongs to the provided input class we return 1. if not 0 is returned

    if g > 0:
        return g,1
    else:
        return g,0

    
start_test()

